#include "sce.h"
#include <iostream>
#include <random>
using namespace Eigen;
using namespace std;

double normpdf(double x){double y=pow(2*M_PI,-0.5)*exp(-pow(x,2)/2);return (y);}
double normkernelf(VectorXd t, VectorXd dF, double ti, double h);
double normkernelfp(VectorXd t, VectorXd dF, double ti, double h);

class AFT_PIC{
private:
    long njump_; long ncount_;
    VectorXd Z_;
    VectorXd beta0_;VectorXd t0_;VectorXd F0_;
    VectorXd dF_; bool Fcov_;
    VectorXd EY_; VectorXi Lline_;VectorXi Rline_;
    
    
    VectorXd fL_; VectorXd fR_; VectorXd fpfsub_; VectorXd FL_;VectorXd FR_;
    VectorXd f_;VectorXd fpf_; VectorXd G_; MatrixXd GX_;
    MatrixXd Xb_; MatrixXd X_var;
    MatrixXd hstar_; VectorXd inthF_; MatrixXd A_; VectorXd Y_;
    VectorXd Pnl; MatrixXd Pnl2;
    
    double SE_; double SE0_;
    double Fdiff_=1; double dnormbeta_=1; double dSE_;
    const double epsilonbeta=1E-4;
    const double epsilonF=1E-4;
    const int maxiter=200; const int maxiterhyb=100;
    const double epsilonSE=1E-2;
    int iter;
    
    
    //Update F0(t0) at new time points t
    VectorXd updateF(VectorXd t0, VectorXd F0, VectorXd t){
        long njump0=t0.size();
        long njump=t.size();
        long tj;
        VectorXd F(njump);
        for (int i=0;i<njump;++i){
            tj=0;
            while (tj<njump0){
                if (t(i)>=t0(tj)) tj=tj+1;
                else break;
            }
            if (tj>0) F(i)=F0(tj-1);else F(i)=0;
        }
        return(F);
    }
    
    //F initiation
    void Finit(){
        // all the exact failure times
        int d=censor_.sum();
        VectorXd Lt(d);
        long line=0;
        for (int i=0;i<n_;++i){
            if (L_(i)>0){eL_(i)=log(L_(i))-X_.row(i)*beta_;}
            if (R_(i)>0){eR_(i)=log(R_(i))-X_.row(i)*beta_;}
            if (censor_(i)==1) {Lt(line)=eL_(i); line=line+1;}
        }
        t_=VecSortUniq(Lt);
        njump_=t_.size();
        F0_.resize(njump_);
        F0_.setZero();
        F_.resize(njump_);
        for (int i=0;i<njump_;++i) F_(i)=((double)i+1)/((double)njump_+1);
    }
    
    //beta initiation
    void betainit(){
        beta_.resize(p_);beta_.setZero();
        beta0_.resize(p_);beta0_.setOnes();
    }
    
    // solve F using self consistency algorithm
    void solveF(VectorXd Z){
        for (int i=0;i<n_;++i){
            if (L_(i)>0){eL_(i)=log(L_(i))-X_.row(i)*beta_;}
            if (R_(i)>0){eR_(i)=log(R_(i))-X_.row(i)*beta_;}
        }
        t0_=t_; F0_=F_;
        sce sceF(eL_,eR_,censor_,censori_,Z,t0_,F0_);
        sceF.fsolve();
        Fcov_=sceF.cov_;
        t_=sceF.t_;F_=sceF.F_;njump_=F_.size();
        Lline_=sceF.Lline_;Rline_=sceF.Rline_;
    }
    // solve F using one step update
    void onestepF(VectorXd Z){
        for (int i=0;i<n_;++i){
            if (L_(i)>0){eL_(i)=log(L_(i))-X_.row(i)*beta_;}
            if (R_(i)>0){eR_(i)=log(R_(i))-X_.row(i)*beta_;}
        }
        t0_=t_; F0_=F_;
        sce sceF(eL_,eR_,censor_,censori_,Z,t0_,F0_);
        sceF.fonestep();
        t_=sceF.t_;F_=sceF.F_;njump_=F_.size();
        Lline_=sceF.Lline_;Rline_=sceF.Rline_;
    }
    
    //EY calculation
    void EY(){
        EY_=X_*beta_; dF_=F_;
        for (int i=1;i<njump_;++i){dF_(i)=F_(i)-F_(i-1);}
        for (int i=0;i<n_;++i){
            if (censor_(i)==1){EY_(i)=EY_(i)+eL_(i);}
            else{
                if ((censori_(i)!=1)&(Lline_(i)==njump_-1)){//L>t(njump)
                    if (censori_(i)==3){EY_(i)=EY_(i)+eL_(i);}
                    else {EY_(i)=EY_(i)+(eL_(i)+eR_(i))/2;}
                } else if ((censori_(i)!=3)&(Rline_(i)==-1)){//R<t1
                    if (censori_(i)==3){EY_(i)=EY_(i)+eR_(i);}
                    else {EY_(i)=EY_(i)+(eL_(i)+eR_(i))/2;}
                } else{
                    double FR; double FL; int Ll; int Rl;
                    if ((censori_(i)==1)|(Lline_(i)==-1)){FL=0;Ll=-1;}
                    else {FL=F_(Lline_(i));Ll=int (Lline_(i));}
                    if (censori_(i)==3){FR=F_(njump_-1);Rl=int (njump_)-1;}
                    else if(Rline_(i)==-1){FR=0;Rl=-1;}
                    else {FR=F_(Rline_(i));Rl=int (Rline_(i));}
                    if (FL==FR){EY_(i)=EY_(i)+(eL_(i)+eR_(i))/2;}
                    else {for (int j=(Ll+1);j<=Rl;++j){EY_(i)=EY_(i)+t_(j)*dF_(j)/(FR-FL);}}
                }
            }
        }
    }
    
    //Least Square
    void LS(VectorXd Z){
        double Ybar; double nZ=Z.sum();
        Ybar=Z.transpose()*EY_; Ybar=Ybar/nZ;
        VectorXd Xbar; Xbar=Z.transpose()*X_;
        Xbar=Xbar/nZ;
        VectorXd betanum(p_);betanum.setZero();
        MatrixXd betaden(p_,p_);betaden.setZero();
        MatrixXd XXbar(n_,p_); XXbar.setZero();
        MatrixXd XbarX2(p_,p_);XbarX2.setZero();
        for (int i=0;i<n_;++i){
            XXbar.row(i)=X_.row(i)-Xbar.transpose();
            XbarX2=Z(i)*(XXbar.row(i).transpose()*XXbar.row(i));
            betaden=betaden+XbarX2;
            for (int j=0;j<p_;++j){betanum(j)=betanum(j)+Z(i)*(X_(i,j)-Xbar(j))*(EY_(i)-Ybar);}
        }
        beta_=betaden.inverse()*betanum;
        double alpha=Ybar-Xbar.transpose()*beta_;
        SE0_=SE_;SE_=0;
        for (int i=0;i<n_;++i) SE_=SE_+Z_(i)*pow(EY_(i)-alpha-XXbar.row(i)*beta_,2)/nZ;
    }
    
    void diffcal(){
        long njump0=t0_.size();Fdiff_=0;
        VectorXd tall (njump0+njump_);
        tall.segment(0,njump0)=t0_;tall.segment(njump0,njump_)=t_;
        VectorXd t=VecSortUniq(tall);// declared at sce.h
        long njump=t.size();
        int linet=0;int linet0=0;
        double Fi; double F0i;
        for (int i=0;i<njump-1;++i){
            // find the smallest linet s.t. t_(linet)>t(i), then the F value at t(i) is F_(linet-1)
            while (linet<njump_){ if (t_(linet)<=t(i)) linet++; else break;}
            while (linet0<njump0){ if (t0_(linet0)<=t(i)) linet0++; else break;}
            if (linet==0) Fi=0; else Fi=F_(linet-1);
            if (linet0==0) F0i=0; else F0i=F0_(linet0-1);
            Fdiff_=Fdiff_+abs((Fi-F0i)*(t(i+1)-t(i)));
        }
        Fdiff_=Fdiff_/(t(njump-1)-t(0));
        VectorXd betadiff(p_);
        betadiff=beta_-beta0_;
        dnormbeta_=betadiff.norm()/beta_.norm();
        dSE_=abs(SE_-SE0_);
    }
    
    //Efficient Estimation
    void LRline(VectorXd eL,VectorXd eR){
        ncount_=n_; FL_.setZero();FR_.setZero(); int t;
        for (int i=0; i<n_;++i){
            if (censor_(i)==1) {
                t=0;
                while (t<njump_){
                    if (eL(i)>=t_(t)) t=t+1; else break;
                }
                Lline_(i)=t-1;
            } else{
                t=0; while (t<njump_){
                    if (eL(i)>=t_(t)) t=t+1; else break;
                }
                Lline_(i)=t-1;
                t=0; while (t<njump_){
                    if (eR(i)>=t_(t)) t=t+1; else break;
                }
                Rline_(i)=t-1;
                if (censori_(i)==1) Lline_(i)=-1;
                if (censori_(i)==3) Rline_(i)=int (njump_)-1;
                if (Lline_(i)<0) FL_(i)=0; else FL_(i)=F_(Lline_(i));
                if (Rline_(i)<0) FR_(i)=0; else FR_(i)=F_(Rline_(i));
                if (FL_(i)==FR_(i)) ncount_=ncount_-1;
            }
        }
    }
    
    double getsigma(VectorXd x, VectorXd w){
        double xstd=0; double x25=0; double x75=0; long n=x.size();
        double xmean=x.transpose()*w; double sumw=0;
        for (int i=0;i<n;++i){
            sumw=sumw+w(i);
            if (sumw<=0.25) x25=x(i);
            if (sumw<=0.75) x75=x(i);
            xstd=xstd+pow(x(i)-xmean,2)*w(i)*double(n)/(double(n)-1.0);
        }
        double sigma=min(pow(xstd,0.5),(x75-x25)/1.34);
        return(sigma);
    }
    
    void kernelsub(VectorXd eL, VectorXd eR, double hf, double hfp){
        for (int i=0;i<n_;++i){
            fL_(i)=normkernelf(t_, dF_, eL(i),hf); fR_(i)=normkernelf(t_, dF_, eR(i),hf);
            if (censor_(i)==1){
                double fp=normkernelfp(t_, dF_, eL(i),hfp);
                fpfsub_(i)=fp/fL_(i);
                if (fpfsub_(i)>10) fpfsub_(i)=10; else if (fpfsub_(i)<-10) fpfsub_(i)=-10;
            }
            else if (censori_(i)==1) fL_(i)=0;
            else if (censori_(i)==3) fR_(i)=0;
        }
    }
    
    void kernelG(VectorXd eL, double hf, double hfp, double hG){
        f_.resize(njump_);fpf_.resize(njump_);
        for (int j=0;j<njump_;++j){
            f_(j)=normkernelf(t_, dF_, t_(j),hf);
            double fp=normkernelfp(t_, dF_, t_(j),hfp);
            fpf_(j)=fp/f_(j);
            if (fpf_(j)>10) fpf_(j)=10; else if (fpf_(j)<-10) fpf_(j)=-10;
        }
        G_.resize(njump_);GX_.resize(njump_,p_);
        G_.setZero();GX_.setZero();
        for (int j=0;j<njump_;++j){
            for (int i=0;i<n_;++i){
                if (censor_(i)==1){
                    G_(j)=G_(j)+normpdf((eL(i)-t_(j))/hG)/hG/f_(j)/double(ncount_);
                    GX_.row(j)=GX_.row(j)+normpdf((eL(i)-t_(j))/hG)/hG/f_(j)*Xb_.row(i)/double(ncount_);
                }
            }
        }
    }
    
    void solveeff(VectorXd beta, double hG, double hf, double hfp){
        LRline(eL_,eR_); kernelsub(eL_,eR_,hf,hfp); kernelG(eL_, hf, hfp, hG);
        
        // solve equation to give hstar
        hstar_.resize(njump_,p_);
        A_.resize(njump_,njump_);Y_.resize(njump_);
        for (int Xk=0;Xk<p_;++Xk){
            A_.setZero();Y_.setZero();
            for (int j=0;j<njump_;++j){
                A_(j,j)=A_(j,j)-G_(j); Y_(j)=Y_(j)+GX_(j,Xk)*fpf_(j);
                for (int l=0;l<njump_;++l) A_(j,l)=A_(j,l)+G_(j)*dF_(l);
                for (int i=0;i<n_;++i){
                    if (censor_(i)==0){
                        if (FR_(i)-FL_(i)>0){
                            double LRj=0;
                            if ((Lline_(i)<j)&(Rline_(i)>=j)) LRj=1/(FR_(i)-FL_(i));
                            Y_(j)=Y_(j)+Xb_(i,Xk)*(fR_(i)-fL_(i))/(FR_(i)-FL_(i))*(LRj)/double(ncount_);
                            for (int l=0;l<njump_;++l){
                                double LRl=0;
                                if ((Lline_(i)<l)&(Rline_(i)>=l)) LRl=1/(FR_(i)-FL_(i));
                                A_(j,l)=A_(j,l)+dF_(l)*(1-LRl)*(LRj)/double(ncount_);
                            }
                        }
                    }
                }
            }
            hstar_.col(Xk)=A_.inverse()*Y_;
        }
        inthF_.resize(p_);inthF_.setZero();
        for (int j=0;j<njump_;++j) inthF_=inthF_+hstar_.row(j).transpose()*dF_(j);
        
        // get the efficient score
        Pnl.resize(p_); Pnl2.resize(p_,p_); Pnl.setZero(); Pnl2.setZero();
        for (int i=0;i<n_;++i){
            if (censor_(i)==1){
                VectorXd a=inthF_-(hstar_.row(Lline_(i))).transpose();
                a=a-fpfsub_(i)*(Xb_.row(i)).transpose();
                Pnl=Pnl+a/double(ncount_);
                Pnl2=Pnl2+a*a.transpose()/(double)ncount_;
            } else{
                if (FR_(i)-FL_(i)>0){
                    VectorXd b=inthF_-(fR_(i)-fL_(i))*(Xb_.row(i)).transpose()/(FR_(i)-FL_(i));
                    for (int j=Lline_(i)+1;j<=Rline_(i);++j) b=b-hstar_.row(j).transpose()*dF_(j)/(FR_(i)-FL_(i));
                    Pnl=Pnl+b/(double)ncount_;
                    Pnl2=Pnl2+b*b.transpose()/(double)ncount_;
                }
            }
        }
        betaeff_=beta+Pnl2.inverse()*Pnl;
    }
    
public:
    long n_; long p_;
    VectorXd t_, F_;
    VectorXd beta_, betaeff_;
    VectorXd betaSE_,betaeffSE_;
    MatrixXd betaCI_,betaeffCI_;
    VectorXd eL_; VectorXd eR_;MatrixXd X_;
    VectorXd L_; VectorXd R_;
    VectorXi censor_; VectorXi censori_;// 1 for left, 2 for interval, 3 for right
    bool coviter_;
    bool covSE_;
    
    AFT_PIC() = default;
    AFT_PIC(VectorXd L,VectorXd R, MatrixXd X, VectorXi censor, VectorXi censori){
        n_=X.rows();p_=X.cols();
        X_=X;censor_=censor;censori_=censori;Z_.resize(n_);Z_.setOnes();
        L_=L;R_=R;
        eL_.resize(n_); eR_.resize(n_);
    }
    AFT_PIC(VectorXd L,VectorXd R, MatrixXd X, VectorXi censor, VectorXi censori, VectorXd Z){
        n_=X.rows();p_=X.cols();
        X_=X;censor_=censor;censori_=censori;Z_=Z;
        L_=L;R_=R;
        eL_.resize(n_); eR_.resize(n_);
    }
    void BJ_est(){
        iter=0;betainit();Finit();
        while ((((dnormbeta_>epsilonbeta)|(Fdiff_>epsilonF)))&(iter<maxiterhyb)){
            iter=iter+1;beta0_=beta_;
            onestepF(Z_);
            EY();
            LS(Z_);
            diffcal();
            //cout << "In the "<<iter<<" iteration, beta = "<<beta_<<endl;
        }
        coviter_=(iter<maxiterhyb);
        if (!coviter_){
            iter=0;
            while ((((dnormbeta_>epsilonbeta)|(Fdiff_>epsilonF)))&(iter<maxiter)){
                iter=iter+1;beta0_=beta_;
                solveF(Z_);
                EY();
                LS(Z_);
                diffcal();
                //cout << "In the "<<iter<<" iteration, beta = "<<beta_<<endl;
            }
            coviter_=(iter<maxiter);
        }
        covSE_=(dSE_<epsilonSE);
        if (!coviter_) beta_=(beta_+beta0_)/2.0;
    }
    void BJ_boot(long nrep, default_random_engine & generator){
        uniform_real_distribution<double> bootdist(0,(double)n_);
        VectorXd L(n_); VectorXd R(n_); VectorXi censor(n_); VectorXi censori(n_); MatrixXd X(n_,p_);
        MatrixXd varres(nrep,p_);
        int rep=0;int trynum=0;
        while(rep<nrep){
            for (int i=0;i<n_;++i){
                double subid = bootdist(generator);
                L(i) = L_(subid);
                R(i) = R_(subid);
                censor(i) = censor_(subid);
                censori(i) = censori_(subid);
                X.row(i) = X_.row(subid);
            }
            AFT_PIC AFTZ(L,R,X,censor,censori);
            AFTZ.BJ_est();trynum++;
            bool cov=AFTZ.covSE_;
            if (cov){
                varres.row(rep)=AFTZ.beta_;
                ++rep;
            }
        }
        betaSE_.resize(p_);betaSE_.setZero();
        betaCI_.resize(p_,2);betaCI_.setZero();
        for (int j=0;j<p_;++j){
            double barX=(varres.col(j)).mean();
            for (int i=0;i<nrep;++i){betaSE_(j)=betaSE_(j)+pow(varres(i,j)-barX,2.0)/(double(nrep)-1.0);}
            betaSE_(j)=pow(betaSE_(j),0.5);
            VectorXd resj = VecSort(varres.col(j));
            betaCI_(j,0) = resj((nrep*0.025-1));
            betaCI_(j,1) = resj((nrep*0.975-1));
        }
    }
    void Onestep_est(){
        BJ_est();
        Xb_.resize(n_,p_); FL_.resize(n_); FR_.resize(n_);
        fL_.resize(n_); fR_.resize(n_); fpfsub_.resize(n_);
        
        VectorXd Z(n_); Z.setOnes(); VectorXd meanX=Z.transpose()*X_/(double)n_;
        X_var = X_.transpose()*X_/(double)n_-meanX*meanX.transpose();
        for (int i=0;i<n_;++i) {Xb_.row(i)=X_.row(i)-meanX.transpose();}
        for (int j=0;j<p_;++j){
            Xb_.col(j) = Xb_.col(j)/sqrt(X_var(j,j));
            beta_(j) = beta_(j)*sqrt(X_var(j,j));
        }
        
        //optimal kernel for f and f' : weighted sd
        int nn=censor_.sum(); VectorXd eLL(nn); VectorXd w(nn); int line=0;
        for (int i=0;i<n_;++i) {if (censor_(i)==1){eLL(line)=eL_(i); w(line)=1/double(nn); line=line+1;}}
        eLL=VecSort(eLL); double sigma=getsigma(eLL,w);
        
        double hG=sigma*pow(double(nn)*(3.0/4.0),-0.2);
        double hf=sigma*pow(double(nn)*(3.0/4.0),-0.2);
        double hfp=sigma*pow(double(nn)*(5.0/4.0),-double(1)/double(7));
        
        solveeff(beta_, hG, hf, hfp);
        
        for (int j=0;j<p_;++j){
            betaeff_(j)=betaeff_(j)/sqrt(X_var(j,j));
            beta_(j) = beta_(j)/sqrt(X_var(j,j));
        }
    }
    void Onestep_boot(long nrep, default_random_engine & generator){
        uniform_real_distribution<double> bootdist(0,(double)n_);
        VectorXd L(n_); VectorXd R(n_); VectorXi censor(n_); VectorXi censori(n_); MatrixXd X(n_,p_);
        MatrixXd varres(nrep,p_);
        int rep=0;int trynum=0;
        while(rep<nrep){
            for (int i=0;i<n_;++i){
                double subid = bootdist(generator);
                L(i) = L_(subid);
                R(i) = R_(subid);
                censor(i) = censor_(subid);
                censori(i) = censori_(subid);
                X.row(i) = X_.row(subid);
            }
            AFT_PIC AFTZ(L,R,X,censor,censori);
            AFTZ.Onestep_est();trynum++;
            bool cov=AFTZ.covSE_;
            if (cov){
                varres.row(rep)=AFTZ.betaeff_;
                ++rep;
            }
        }
        betaeffSE_.resize(p_);betaeffSE_.setZero();
        betaeffCI_.resize(p_,2);betaeffCI_.setZero();
        for (int j=0;j<p_;++j){
            double barX=(varres.col(j)).mean();
            for (int i=0;i<nrep;++i){betaeffSE_(j)=betaeffSE_(j)+pow(varres(i,j)-barX,2.0)/(double(nrep)-1.0);}
            betaeffSE_(j)=pow(betaeffSE_(j),0.5);
        }
    }
};

double normkernelf(VectorXd t, VectorXd dF, double ti, double h){
    double f=0; long njump=t.size();
    for (int j=0;j<njump;++j) {f=f+normpdf((t(j)-ti)/h)*dF(j)/h;}
    return(f);
}
double normkernelfp(VectorXd t, VectorXd dF, double ti, double h){
    double fp=0; long njump=t.size();
    for (int j=0;j<njump;++j) {fp=fp+normpdf((t(j)-ti)/h)*(t(j)-ti)*dF(j)*pow(h,-3);}
    return(fp);
}
